import logging
from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
    cast,
)

from pcs.common.file import RawFileError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.json import JsonParserException
from pcs.lib.interface.config import ParserErrorException

from . import const
from .config.facade import Facade
from .config.parser import ParserError
from .pam import authenticate_user
from .tools import (
    UserGroupsError,
    get_user_groups,
)


@dataclass(frozen=True)
class AuthUser:
    username: str
    groups: Sequence[str]

    @property
    def is_superuser(self) -> bool:
        return self.username == const.SUPERUSER


class AuthProvider:
    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger
        self._config_file_instance = FileInstance.for_pcs_users_config()

    def _get_facade(self) -> Facade:
        try:
            return cast(Facade, self._config_file_instance.read_to_facade())
        except ParserError as e:
            self._logger.error(
                "Unable to parse file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.msg,
            )
        except JsonParserException:
            self._logger.error(
                "Unable to parse file '%s': not valid json",
                self._config_file_instance.raw_file.metadata.path,
            )
        except ParserErrorException:
            self._logger.error(
                "Unable to parse file '%s'",
                self._config_file_instance.raw_file.metadata.path,
            )
        except RawFileError as e:
            self._logger.error(
                "Unable to read file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.reason,
            )
        return Facade([])

    def _write_facade(self, facade: Facade) -> None:
        try:
            self._config_file_instance.write_facade(facade, can_overwrite=True)
        except RawFileError as e:
            self._logger.error(
                "Action '%s' on file '%s' failed: %s",
                e.action,
                e.metadata.path,
                e.reason,
            )

    def login_user(self, username: str) -> Optional[AuthUser]:
        try:
            groups = get_user_groups(username)
        except UserGroupsError:
            self._logger.error(
                "Unable to determine groups of user '%s'", username
            )
            return None
        if const.ADMIN_GROUP not in groups:
            self._logger.debug(
                "User '%s' is not a member of '%s' group",
                username,
                const.ADMIN_GROUP,
            )
            return None
        self._logger.debug("Successful login by '%s'", username)
        return AuthUser(username=username, groups=tuple(groups))

    def auth_by_token(self, token: str) -> Optional[AuthUser]:
        username = self._get_facade().get_user(token)
        if username is None:
            return None
        return self.login_user(username)

    def auth_by_username_password(
        self, username: str, password: str
    ) -> Optional[AuthUser]:
        if authenticate_user(username, password):
            return self.login_user(username)
        self._logger.info(
            "Failed login by '%s': bad username or password", username
        )
        return None

    def create_token(self, username: str) -> str:
        facade = self._get_facade()
        token = facade.add_user(username)
        self._write_facade(facade)
        return token



import grp
import pwd


class UserGroupsError(Exception):
    pass


def get_user_groups(username: str) -> list[str]:
    try:
        return [
            group.gr_name
            for group in grp.getgrall()
            if username in group.gr_mem
        ] + [
            # `group.gr_mem` does not contain the `username` when `group` is
            # primary for the `username` (the same is in /etc/group). So it is
            # necessary to add the primary group.
            grp.getgrgid(pwd.getpwnam(username).pw_gid).gr_name
        ]
    except KeyError as e:
        raise UserGroupsError from e



import time
from typing import (
    Iterable,
    Optional,
    Sequence,
    cast,
)

from pcs.common.tools import get_unique_uuid
from pcs.lib.interface.config import FacadeInterface

from .types import TokenEntry


class Facade(FacadeInterface):
    def __init__(self, parsed_config: Iterable[TokenEntry]) -> None:
        super().__init__(parsed_config)

    def _set_config(self, config: Iterable[TokenEntry]) -> None:
        super()._set_config(list(config))

    @property
    def config(self) -> Sequence[TokenEntry]:
        return tuple(cast(Iterable[TokenEntry], super().config))

    def _get_entry_by_token(self, token: str) -> Optional[TokenEntry]:
        return {entry.token: entry for entry in self.config}.get(token)

    def get_user(self, token: str) -> Optional[str]:
        entry = self._get_entry_by_token(token)
        if entry is None:
            return None
        return entry.username

    def add_user(self, username: str) -> str:
        new_entry = TokenEntry(
            username=username,
            token=get_unique_uuid(tuple(entry.token for entry in self.config)),
            creation_date=time.strftime("%Y-%m-%d %H:%M:%S %z"),
        )
        self._set_config(list(self.config) + [new_entry])
        return new_entry.token
